"""The HEPi package aims to automize cluster computations for parameter scans with the option to produce plots."""
from .output import *
from .input import *
from .results import *
from .util import *
from .plot import *
from .run import *
from .load import *

import pkg_resources as pkg  # part of setuptools
#import json
#import requests
#from urllib.request import urlopen

package = "hepi"

try:
    version = pkg.require(package)[0].version
except pkg.DistributionNotFound:
    version = "dirty"

__version__ = version


import copy
import json
from re import I
from uncertainties import ufloat
from hepi.input import Input, order_to_string, xsec_to_order
from hepi.util import LD2DL, DL2DF


def load(f, dimensions=1):
    """
    Load xsec data from json in to something that works within hepi's plotting.

    Args:
        f : readable object, e.g. `open(filepath:str)`.
        dimensions (int) : 1 or 2 currently supported.
    """
    dict = json.load(f)

    inpu = Input(
        xsec_to_order(dict["order"]),
        float(dict["Ecom [GeV]"]),
        -1,
        -1,
        slha=dict["process_latex"],
        pdf_lo=dict["PDF set"],
        pdf_nlo=dict["PDF set"],
        update=False,
    )
    inpu.runner = dict["tool"]
    dat = []
    params = dict["parameters"]
    if dimensions == 2:
        for k in dict["data"]:
            for l in dict["data"][k]:
                dicd = copy.copy(inpu.__dict__)
                for p1 in params[0]:
                    dicd[p1] = float(k)
                for p2 in params[1]:
                    dicd[p2] = float(l)
                if "unc_pb" in dict["data"][k][l]:
                    dicd[order_to_string(inpu.order)] = ufloat(
                        dict["data"][k][l]["xsec_pb"],
                        dict["data"][k][l]["unc_pb"])
                elif "unc_down_pb" in dict["data"][k][l] and "unc_up_pb" in dict["data"][k][l]:
                    dicd[order_to_string(inpu.order)+ "_NOERR"] = dict["data"][k][l]["xsec_pb"]
                    dicd[order_to_string(inpu.order)+ "_COMBINED"] = ufloat(
                        dict["data"][k][l]["xsec_pb"] + (dict["data"][k][l]["unc_up_pb"] + dict["data"][k][l]["unc_down_pb"]) / 2,
                        (dict["data"][k][l]["unc_up_pb"] - dict["data"][k][l]["unc_down_pb"]) / 2)
                else:
                    raise ValueError("No uncertainty found in data.")
                dat.append(dicd)
    if dimensions == 1:
        for k in dict["data"]:
            dicd = copy.copy(inpu.__dict__)
            for p1 in params[0]:
                dicd[p1] = float(k)
            if "unc_pb" in dict["data"][k]:
                dicd[order_to_string(inpu.order)] = ufloat(
                    dict["data"][k]["xsec_pb"], dict["data"][k]["unc_pb"])
            elif "unc_down_pb" in dict["data"][k] and "unc_up_pb" in dict["data"][k]:
                dicd[order_to_string(inpu.order)+ "_NOERR"] = dict["data"][k]["xsec_pb"]
                dicd[order_to_string(inpu.order)+ "_COMBINED"] = ufloat(
                    dict["data"][k]["xsec_pb"] + (dict["data"][k]["unc_up_pb"] + dict["data"][k]["unc_down_pb"]) / 2,
                    (dict["data"][k]["unc_up_pb"] - dict["data"][k]["unc_down_pb"]) / 2)
            else:
                raise ValueError("No uncertainty found in data.")
            dat.append(dicd)
    return DL2DF(LD2DL(dat, actual_dict=True))



import os
import pkgutil
import shutil
import stat
from string import Template
import subprocess
import warnings
from hepi.input import Input, Order, is_gluino, is_slepton, is_squark, is_weakino
from hepi.results import Result
from hepi.run import RunParam, Runner
from uncertainties import ufloat


class ProspinoResult(Result):
    """
    Prospino Result Data.
    """


class ProspinoRunner(Runner):
    # TODO add leptoqaurk, stop and sbottom cases
    weakino_map = {
        1000022: 1,
        1000023: 2,
        1000025: 3,
        1000035: 4,
        1000024: 5,
        1000037: 6,
        -1000024: 7,
        -1000037: 8,
    }
    squark_map = {
        1000005: -5,
        1000004: -4,
        1000003: -3,
        1000001: -2,
        1000002: -1,
        2000002: 1,
        2000001: 2,
        2000003: 3,
        2000004: 4,
        2000005: 5,
    }
    slepton_map = {
        (1000011, 1000011): 1,
        (2000011, 2000011): 2,
        (1000012, 1000012): 3,
        (-1000011, 1000012): 4,
        (1000011, 1000012): 5,
        (1000015, 1000015): 6,
        (2000015, 2000015): 7,
        (1000015, 2000015): 8,
        (1000016, 1000016): 9,
        (-1000015, 1000016): 10,
        (1000015, 1000016): 11,
        (-2000015, 1000016): 12,
        (2000015, 1000016): 13,
    }

    def _get_ps_proc(self, p: Input):
        if p.has_squark() and p.has_weakino():
            sq = p.particle1 if is_squark(p.particle1) else p.particle2
            weakino = p.particle2 if is_squark(p.particle1) else p.particle1
            return 'ns', self.weakino_map[weakino], 1, self.squark_map[abs(
                sq)], 1
        if p.has_gluino() and p.has_weakino():
            return 'ng', self.weakino_map[p.particle1 if is_gluino(p.particle2
                                                                   ) else p.
                                          particle2], 1, 1, 1
        if p.has_gluino() and p.has_squark():
            return 'sg', 1, 1, self.squark_map[abs(
                p.particle1 if is_gluino(p.particle2) else p.particle2)], 1
        if is_weakino(p.particle1) and is_weakino(p.particle2):
            return 'nn', self.weakino_map[p.particle1], self.weakino_map[
                p.particle2], 1, 1
        if is_slepton(p.particle1) and is_slepton(p.particle2):
            return 'll', self.slepton_map[(p.particle1, p.particle2)], 1, 1, 1
        if is_gluino(p.particle1) and is_gluino(p.particle2):
            return 'gg', 1, 1, 1, 1
        if is_squark(p.particle1) and is_squark(p.particle2):
            if p.particle1 > 0 and p.particle2 > 0:
                return 'ss', 1, 1, self.squark_map[abs(
                    p.particle1)], self.squark_map[abs(p.particle2)]
            elif (p.particle1 > 0
                  and p.particle2 < 0) or (p.particle1 < 0
                                           and p.particle2 > 0):
                s = p.particle1 if p.particle1 > p.particle2 else p.particle2
                b = p.particle2 if p.particle1 > p.particle2 else p.particle1
                return 'sb', 1, 1, self.squark_map[abs(s)], self.squark_map[
                    abs(b)]
        return "UNKNOWN_PROCESS_OR_UNIMPLEMENTED_PROCESS"

    def _get_ps_input(self, p: Input) -> dict:
        d = {}
        d["ps_inlo"] = int(p.order)
        d["ps_final_state_in"], d["ps_ipart1_in"], d["ps_ipart2_in"], d[
            "ps_isquark1_in"], d["ps_isquark2_in"] = self._get_ps_proc(p)
        return d

    def orders(self) -> Order:
        return [Order.LO, Order.NLO]

    def _check_input(self, p: Input, **kwargs) -> bool:
        """Checks input parameter for compatibility with Prospino"""
        if p.mu_f != 1. or p.mu_r != 1.:
            warnings.warn(
                "Prospino2 does not support varying the scales manually.")
            return False
        if p.pdf_lo != "cteq6l1" or p.pdf_nlo != "cteq66":
            warnings.warn(
                "Prospino2 does not support all pdfs (CTEQ6L1 and CTEQ66 allowed defaults)."
            )
            return False
        return True

    def _is_valid(self, file: str, p: Input, d) -> bool:
        return super()._is_valid(file, p, d)

    def _parse_file(self, file: str) -> Result:
        ret = []
        with open(file) as output:
            for line in output:
                if line.startswith("nn") or line.startswith(
                        "ng") or line.startswith("ns") or line.startswith(
                            "sg") or line.startswith("ll") or line.startswith(
                                "gg") or line.startswith(
                                    "ss") or line.startswith(
                                        "sb"):  # TODO generalize
                    for s in line[2:].split():
                        ret.append(float(s))
        return ProspinoResult(
            ufloat(ret[8], ret[8] * ret[9]),
            ufloat(ret[10], ret[10] * ret[11]) if ret[10] != 0. else None,
            None)

    def _prepare(self, p: Input, **kwargs) -> RunParam:
        rp = super()._prepare(p, **kwargs)
        if not rp.skip:
            d = p.__dict__
            data = pkgutil.get_data(
                __name__, "prospino_main.f90_template").decode('utf-8')
            src = Template(data)
            #compute dependent pieces for template
            d["ps_inlo"] = int(p.order)
            for k, v in self._get_ps_input(p).items():
                d[k] = v
            result = src.substitute(d)
            #open(rp.in_file, "w").write(result)
            open(rp.out_file, "w").write(result + "\n\n")
            rdir = self.get_output_dir() + rp.name + ".rdir"
            if os.path.exists(rdir) and os.path.isdir(rdir):
                shutil.rmtree(rdir)
            shutil.copytree(self.get_path(), rdir)
            open(rdir + "/prospino_main.f90", "w").write(result)
            # compile
            subprocess.Popen("cd " + rdir + " && make",
                             shell=True,
                             stdout=subprocess.DEVNULL).wait()

            open(rp.execute, "w").write(
                "#!/bin/sh\n" +
                "H=$PWD && cd {rdir}&& ./prospino_2.run > tmp.out && cd $H  && cat {rdir}/tmp.out  >> {out} && cat {rdir}/prospino.dat>> {out} && rm -rf {rdir}"
                .format(rdir=rdir, out=rp.out_file))
            st = os.stat(rp.execute)
            os.chmod(rp.execute, st.st_mode | stat.S_IEXEC)

            sname = d['slha']
            shutil.copy(self.get_output_dir() + sname,
                        rdir + "/prospino.in.les_houches")
            with open(self.get_output_dir() + sname, 'r') as f:
                open(rp.out_file, "a").write(f.read() + "\n\n")
        return rp


# Legacy
default_prospino_runner = ProspinoRunner("~/git/Prospino2/")
"""Default Prospino Runner to provide backward compatibility"""
run = default_prospino_runner.run
set_path = default_prospino_runner.set_path
get_path = default_prospino_runner.get_path


"""Collection of utility functions for the :mod:`hepi` package."""
import json
from typing import List, Tuple
import numpy as np
import hashlib
from particle.converters.bimap import DirectionalMaps
from particle import PDGID
from distutils.version import StrictVersion

import pandas as pd
import warnings
from smpl import interpolate as ip


class DictData:

    def __str__(self):
        """Returns attributes as dict as string"""
        return str(self.__dict__)


def LD2DL(l: List, actual_dict=False) -> dict:
    """
    Convert a list of objects into a dictionary of lists.

    The values of each object are first converted to a `dict` through the `__dict__` attribute.

    Args:
        l (List) : list of objects.
        actual_dict (bool) : objects are already dicts

    Returns:
        dict : dictionary of numpy arrays.

    Examples:
        >>> class Param:
        ...      def __init__(self,a,b,c):
        ...         self.a = a
        ...         self.b = b
        ...         self.c = c
        >>> LD2DL([ Param(1,2,3), Param(4,5,6) , Param(7,8,9) ])
        {'a': array([1, 4, 7]), 'b': array([2, 5, 8]), 'c': array([3, 6, 9])}
    """
    # Check l[0] keys in all dictionaries.
    for m in l:
        md = m if actual_dict else m.__dict__
        for k in l[0] if actual_dict else l[0].__dict__:
            assert k in md
    # switch them
    return {
        k: np.array([dic[k] if actual_dict else dic.__dict__[k] for dic in l])
        for k in (l[0] if actual_dict else l[0].__dict__)
    }


def DL2DF(ld: dict) -> pd.DataFrame:
    """
    Convert a `dict` of `list`s to a `pandas.DataFrame`.
    """
    return pd.DataFrame.from_dict(ld)


PDG2LaTeXNameMap, LaTeX2PDGNameMap = DirectionalMaps("PDGID",
                                                     "LaTexName",
                                                     converters=(PDGID, str))

PDG2Name2IDMap, PDGID2NameMap = DirectionalMaps("PDGName",
                                                "PDGID",
                                                converters=(str, PDGID))


def get_name(pid: int) -> str:
    """
    Get the latex name of a particle.

    Args:
        pid (int) : PDG Monte Carlo identifier for the particle.

    Returns:
        str: Latex name.

    Examples:
        >>> get_name(21)
        'g'
        >>> get_name(1000022)
        '\\\\tilde{\\\\chi}_{1}^{0}'
    """
    global PDG2LaTeXNameMap
    pdgid = PDG2LaTeXNameMap[pid]
    return pdgid


def get_LR_partner(pid: int) -> Tuple[int, int]:
    """Transforms a PDG id to it's left-right partner.

    Args:
        pid (int) : PDG Monte Carlo identifier for the particle.

    Returns:
        tuple : First int is -1 for Left and 1 for Right. Second int is the PDG id.

    Examples:
        >>> get_LR_partner(1000002)
        (-1, 2000002)
    """
    n = PDGID2NameMap[pid]
    if "L" in n:
        n = n.replace("L", "R")
        return -1, int(PDG2Name2IDMap[n])
    if "R" in n:
        n = n.replace("R", "L")
        return 1, int(PDG2Name2IDMap[n])
    if "1" in n:
        n = n.replace("1", "2")
        return -1, int(PDG2Name2IDMap[n])
    if "2" in n:
        n = n.replace("2", "1")
        return 1, int(PDG2Name2IDMap[n])
    return None


def namehash(n: any) -> str:
    """
    Creates a sha256 hash from the objects string representation.

    Args:
        n (any) : object.

    Returns:
        str: sha256 of object.

    Examples:
        >>> p = {'a':1,'b':2}
        >>> str(p)
        "{'a': 1, 'b': 2}"
        >>> namehash(str(p))
        '3dffaea891e5dbadb390da33bad65f509dd667779330e2720df8165a253462b8'
        >>> namehash(p)
        '3dffaea891e5dbadb390da33bad65f509dd667779330e2720df8165a253462b8'
    """
    m = hashlib.sha256()
    m.update(str(n).encode('utf-8'))
    return m.hexdigest()


def lhapdf_name_to_id(name: str) -> int:
    """
    Converts a LHAPDF name to the sets id.

    Args:
        name (str) : LHAPDF set name.

    Returns:
        int: id of the LHAPDF set.

    Examples:
        >>> lhapdf_name_to_id("CT14lo")
        13200
    """
    try:
        import lhapdf
    except ImportError:
        warnings.warn("LHAPDF python binding not installed? Make sure you set PYTHONPATH correctly (i.e. correct python version).")
        return 0
    if not lhapdf.availablePDFSets():
        warnings.warn("No PDF sets found. Make sure the environment variable LHAPDF_DATA_DIR points to the correct location (.../share/LHAPDF).")
        return 0
    if not name in lhapdf.availablePDFSets():
        warnings.warn("PDF set '" + name + "' not installed?")
        return 0
    return lhapdf.getPDFSet(name).lhapdfID

def lhapdf_id_to_name(lid : int) -> str:
    try:
        import lhapdf
    except ImportError:
        warnings.warn("LHAPDF python binding not installed? Make sure you set PYTHONPATH correctly (i.e. correct python version).")
        return ""
    if not lhapdf.availablePDFSets():
        warnings.warn("No PDF sets found. Make sure the environment variable LHAPDF_DATA_DIR points to the correct location (.../share/LHAPDF).")
        return 0
    for n in lhapdf.availablePDFSets():
        if lhapdf.getPDFSet(n).lhapdfID == lid:
            return n
        
    warnings.warn("PDF set with id " + str(lid) + " unknown/not installed?")
    return "Unknown PDF ID: " + str(lid)

# TODO fix dependent(mu) for new masses

def interpolate_1d(df,x,y,xrange,only_interpolation=True):
    """
    Last key is the value to be interpolated, while the rest are cooridnates.
    
    Args:
        df (pandas.DataFrame): results
    """
    f = ip.interpolate(df[x],df[y])
    a = []
    for xr in xrange:
        c = df.head(1)
        c[x] = xr
        c[y] = f(xr)
        a += [c]
    if only_interpolation:
        return pd.concat(a)
    else:
        return pd.concat([df, *a])
 
def interpolate_2d(df,x,y,z,xrange,yrange,only_interpolation=True,**kwargs):
    """
    Last key is the value to be interpolated, while the rest are cooridnates.
    
    Args:
        df (pandas.DataFrame): results
    """
    f = ip.interpolate(df[x],df[y],df[z],**kwargs)
    a = []
    for i in range(len(xrange)):
        xr = xrange[i]
        yr = yrange[i]
        zr = f(xr,yr)
        c = df.head(1).copy()
        c[x] = xr
        c[y] = yr
        c[z] = zr
        a += [c]
    if only_interpolation:
        return pd.concat(a)
    else:
        return pd.concat([df, *a])



"""Runs MadGraph."""
from typing import List
import subprocess
from string import Template
from hepi.input import Order
from hepi.run import RunParam, Runner
import pkgutil
from .. import Input, Result, get_output_dir
from .result import is_valid, parse_single
import time
import os


class MadGraphRunParams(RunParam):
    """Parameters for MadGraph."""

    def __init__(self, dic, skip=False, madstr=True):
        super().__init__(skip)
        self.dic = dic
        # self.skip = skip
        self.madstr = madstr


class MadGraphRunner(Runner):

    def orders(self) -> List[Order]:
        return [Order.LO, Order.NLO]

    def _check_path(self) -> bool:
        if os.path.exists(os.path.expanduser(self.get_path() +
                                             "/bin/mg5_aMC")):
            self.set_path(self.get_path() + "/bin/mg5_aMC")
            return True
        if self.get_path().endswith("mg5_aMC"):
            return True
        return False

    def _check_input(self, param: Input, **kwargs) -> bool:
        """Checks input parameter for compatibility with Prospino"""
        return True

    def _is_valid(self, file: str, p: Input, d) -> bool:
        if not super()._is_valid(file, p, d):
            return False
        return is_valid(file, p, d)

    def _parse_file(self, file: str) -> Result:
        return parse_single(file)

    def _run(self,
             rps: List[RunParam],
             wait=True,
             parallel=True,
             sleep=0,
             **kwargs):
        # TODO clean up on exit emergency
        skipall = True
        for rp in rps:
            if not rp.skip:
                skipall = False
        template = ""
        if not skipall:
            lo = "&& nice -n 5 {dir}/bin/calculate_xsect LO -f >> {out} " if rps[
                0].dic["order"] == Order.NLO else ""
            template =  \
                'rm -rf {dir} && cp -r ' + rps[0].dic["bdir"] + \
                ' {dir}  && cp {slha} {dir}/Cards/param_card.dat && cp {run} {dir}/Cards/run_card.dat && echo "nb_core = 1" >> {dir}/Cards/amcatnlo_configuration.txt ' + lo+ '&& nice -n 5 {dir}/bin/calculate_xsect -f >> {out}  && rm -rf {dir}'
            print(rps[0].dic["out"])
        if not rps[0].skip:
            mgcom = ''
            if rps[0].madstr:
                mgcom = ' --mode="MadSTR"'
            com = self.get_path() + mgcom + \
                ' --file {in} >> {out} && cp {slha} {bdir}/Cards/param_card.dat && cp {run} {bdir}/Cards/run_card.dat && sed -i \'s/.*= req_acc_FO/ 1 = req_acc_FO/g\' {bdir}/Cards/run_card.dat && echo "automatic_html_opening = False" >> {bdir}/Cards/amcatnlo_configuration.txt && nice -n 5 {bdir}/bin/calculate_xsect -f'
            pp = subprocess.Popen(com.format(**rps[0].dic), shell=True)
            pp.wait()
        # Run commands in parallel
        processes = []

        for rp in rps:
            if not rp.skip:
                command = template.format(**rp.dic)
                # print(command)
                process = subprocess.Popen(command, shell=True)
                processes.append(process)
                if not parallel:
                    process.wait()
                # Forced delay to prevent overloading clusters when registering jobs
                time.sleep(sleep)
        if wait:
            # Collect statuses
            output = [p.wait() for p in processes]
            return output
        return []

    def _prepare(self, p: Input, **kwargs) -> RunParam:
        rp = super()._prepare(p, **kwargs)
        name = rp.name
        if not rp.skip:
            d = p.__dict__
            d["dir"] = get_output_dir() + name + ".dir"
            d["bdir"] = get_output_dir() + name + ".bdir"

            infile = ""
            if p.order is Order.LO:
                infile = "lo.mg"
            elif p.order is Order.NLO:
                infile = "nlo.mg"
            else:
                raise ValueError("Madgraph only supported for LO/NLO.")

            data = pkgutil.get_data(__name__, infile).decode('utf-8')
            src = Template(data)
            result = src.substitute(d)
            open(get_output_dir() + name + ".mg", "w").write(result)
            if not rp.skip:
                open(get_output_dir() + name + ".out",
                     "w").write(result + "\n\n")

            if p.order == Order.LO:
                mgfile = "run_card_no_madstr.dat"
            elif p.order == Order.NLO:
                mgfile = "run_card_with_madstr.dat"
            else:
                raise ValueError("Order must be one of LO/NLO in MadGraph.")

            if "madstr" in kwargs and not kwargs["madstr"]:
                mgfile = "run_card_no_madstr.dat"
                rp.madstr = False
            else:
                rp.madstr = True
            data = pkgutil.get_data(__name__, mgfile).decode('utf-8')

            src = Template(data)
            result = src.substitute(d)
            open(get_output_dir() + name + ".dat", "w").write(result)
            if not rp.skip:
                open(get_output_dir() + name + ".out",
                     "a").write(result + "\n\n")

            sname = d['slha']
            with open(get_output_dir() + sname, 'r') as f:
                #src = Template(f.read())
                #result = src.substitute(d)
                #open(get_input_dir() + sname + ".in", "w").write(result)
                if not rp.skip:
                    open(get_output_dir() + name + ".out",
                         "a").write(f.read() + "\n\n")
            rp.dic = {
                'in': get_output_dir() + name + ".mg",
                'dir': d["dir"],
                'bdir': d["bdir"],
                'run': get_output_dir() + name + ".dat",
                'slha': get_output_dir() + sname,
                'out': get_output_dir() + name + ".out",
                'order': p.order
            }
            rp.out_file = rp.dic['out']
        return rp


# Legacy
default_madgraph_runner = MadGraphRunner("/opt/MG5_aMC_v2_7_0/")
"""Default MadGraph Runner to provide backward compatibility"""
run = default_madgraph_runner.run
set_path = default_madgraph_runner.set_path
get_path = default_madgraph_runner.get_path

# print(sp)
# names = queue(sp)
# run(names)



"""Results and postprocessing for the :mod:`hepi` package."""
from .util import DictData
import numpy as np
from uncertainties import unumpy
from smpl import plot
import warnings
import tqdm
#from pqdm.processes import pqdm as ppqdm
from pqdm.threads import pqdm as tpqdm
import multiprocessing as mp
from smpl.parallel import par

#If the numerical uncertainty times :attr:`required_numerical_uncertainty_factor` is higher than the scale or pdf uncertainty a warning is shown.
required_numerical_uncertainty_factor = 5


def my_parallel(arr, f, n_jobs=None, desc=None):
    """
    Parallel execution of f on each element of args
    """
    n_jobs = n_jobs or mp.cpu_count()
    sa = np.array_split(np.array(arr), len(arr) / n_jobs)
    res = []
    for i in tqdm.tqdm(range(len(sa)), desc=desc):
        res += par(f, sa[i])
    return res


class Result(DictData):
    """
    General result class.

    Attributes:
        LO (:obj:`double`): Leading Order result. Defaults to None.
        NLO (:obj:`double`): Next-to-Leading Order result. Defaults to None.
        NLO_PLUS_NLL (:obj:`double`): Next-to-Leading Order plus Next-to-Leading Logarithm result. Defaults to None.
        K_LO (:obj:`double`): LO divided by LO.
        K_NLO (:obj:`double`): NLO divided by LO result.
        K_NLO_PLUS_NLL (:obj:`double`): NLO+NLL divided by LO.
        K_aNNLO_PLUS_NNLL (:obj:`double`): aNNLO+NNLL divided by LO.
        NLO_PLUS_NLL_OVER_NLO (:obj:`double`): NLO+NLL divided by NLO.
        aNNLO_PLUS_NNLL_OVER_NLO (:obj:`double`): aNNLO+NNLL divided by NLO.
    """

    def __init__(self,
                 lo=None,
                 nlo=None,
                 nlo_plus_nll=None,
                 annlo_plus_nnll=None):
        """
        Sets given and computes dependent ``Attributes``.

        Args:
            lo (:obj:`double`): corresponds to :attr:`LO`.
            nlo (:obj:`double`): corresponds to :attr:`NLO`.
            nlo_plus_nll (:obj:`double`): corresponds to :attr:`NLO_PLUS_NLL`.
            annlo_plus_nnll (:obj:`double`): corresponds to :attr:`aNNLO_PLUS_NNLL`.
        """
        self.LO = lo
        self.NLO = nlo
        self.NLO_PLUS_NLL = nlo_plus_nll
        self.aNNLO_PLUS_NNLL = annlo_plus_nnll
        if lo is not None and lo != 0:
            self.K_LO = lo / lo
        else:
            self.K_LO = None
        if nlo is not None and lo != 0:
            self.K_NLO = nlo / lo
        else:
            self.K_NLO = None
        if nlo_plus_nll is not None and lo != 0:
            self.K_NLO_PLUS_NLL = nlo_plus_nll / lo
        else:
            self.K_NLO_PLUS_NLL = None

        if nlo_plus_nll is not None and nlo != 0:
            self.NLO_PLUS_NLL_OVER_NLO = nlo_plus_nll / nlo
        else:
            self.NLO_PLUS_NLL_OVER_NLO = None

        if annlo_plus_nnll is not None and lo != 0:
            self.K_aNNLO_PLUS_NNLL = annlo_plus_nnll / lo
        else:
            self.K_aNNLO_PLUS_NNLL = None

        if annlo_plus_nnll is not None and nlo != 0:
            self.aNNLO_PLUS_NNLL_OVER_NLO = annlo_plus_nnll / nlo
        else:
            self.aNNLO_PLUS_NNLL_OVER_NLO = None


# TODO detect which errors also for scales
def pdf_errors(li,
               dl,
               ordernames=None,
               confidence_level=90,n_jobs=None):
    """
    Just like `pdf_error` but over a list of ordernames.
    """
    if ordernames is None:
        ordernames = ["LO", "NLO", "aNNLO_PLUS_NNLL"]
    r_dl = dl
    for o in ordernames:
        r_dl = pdf_error(li, r_dl, o, confidence_level=confidence_level,n_jobs=n_jobs)

    return r_dl


def _pdf_error_single(members, i, dl, ordername, confidence_level=90):  
    try:
        import lhapdf
    except ImportError:
        raise RuntimeError("LHAPDF with python bindings needed to compute PDF uncertainties. Make sure you set the PYTHONPATH correctly (i.e. correct python version).")
    if not lhapdf.availablePDFSets():
        raise RuntimeError("No PDF sets found. Make sure the environment variable LHAPDF_DATA_DIR points to the correct location (.../share/LHAPDF).")
    if dl["pdfset_nlo"][i] == 0 and dl["mu_f"][i] == 1.0 and dl["mu_r"][
            i] == 1.0:
        pdfset = lhapdf.getPDFSet(dl["pdf_nlo"][i])
        pdfs = [0.0] * pdfset.size
        ddl = dl[members].drop(columns=["pdfset_nlo", "precision","max_iters"])
        bol = ddl.eq(ddl.iloc[i]).all(axis='columns')
        for j in range(len(dl["pdfset_nlo"])):
            if bol[j]:
                pdfs[dl["pdfset_nlo"][j]] = j

    # lo_unc = pdfset.uncertainty(
    #    [plot.unv(dl["LO"][k]) for k in pdfs], -1)
    #if ordername == "LO":
    #    dl.loc[i, "LO_PDF_CENTRAL"] = plot.unv(dl["LO"][i])
    #    dl.loc[i, "LO_PDF_ERRPLUS"] = 0.0
    #    dl.loc[i, "LO_PDF_ERRMINUS"] = 0.0
    #    dl.loc[i, "LO_PDF_ERRSYM"] = 0.0
    #else:
    #print([float(plot.unv(dl[ordername][k])) for k in pdfs])
        nlo_unc = pdfset.uncertainty(
            [float(plot.unv(dl[ordername][k])) for k in pdfs],
            confidence_level)
    return (i,nlo_unc)


def pdf_error(li, dl, ordername="LO", confidence_level=90, n_jobs=None):
    """
    Computes Parton Density Function (PDF) uncertainties through :func:`lhapdf.set.uncertainty`.

    Args:
        li (:obj:`list` of :class:`Input`): Input list.
        dl (:obj:`dict`): :class:`Result` dictionary with lists per entry.
        ordername (`str`): Name of the order.
        confidence_level (:obj:`double`): Confidence Level for PDF uncertainty

    Returns:
        :obj:`dict`: Modified `dl` with new `ordername_{PDF,PDF_CENTRAL,PDF_ERRPLUS,PDF_ERRMINUS,PDF_ERRSYM}` entries.
            - (`ordername`)_`PDF` contains a symmetrized :mod:`uncertainties` object.
    """
    global required_numerical_uncertainty_factor

    example = li[0]
    members = [
        attr for attr in dir(example)
        if not callable(getattr(example, attr)) and not attr.startswith("__")
    ]

    dl[ordername + "_PDF"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl[ordername + "_PDF_CENTRAL"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl[ordername + "_PDF_ERRPLUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl[ordername + "_PDF_ERRMINUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl[ordername + "_PDF_ERRSYM"] = np.array([None] * len(dl["pdfset_nlo"]))

    args = [{
        "members": members,
        "i": i,
        "dl": dl,
        "ordername": ordername,
        "confidence_level": confidence_level
    } for i in range(len(dl["pdfset_nlo"])) if dl["pdfset_nlo"][i] == 0
            and dl["mu_f"][i] == 1.0 and dl["mu_r"][i] == 1.0]
    ret = tpqdm(args,
                _pdf_error_single,
                n_jobs=n_jobs if n_jobs is not None else mp.cpu_count(),
                argument_type='kwargs',
                desc="PDF uncertainty @ " + ordername)
    for i, nlo_unc in ret:
        dl.loc[i, ordername + "_PDF_CENTRAL"] = nlo_unc.central
        dl.loc[i, ordername + "_PDF_ERRPLUS"] = nlo_unc.errplus
        dl.loc[i, ordername + "_PDF_ERRMINUS"] = -nlo_unc.errminus
        dl.loc[i, ordername + "_PDF_ERRSYM"] = nlo_unc.errsymm
        #TODO error sym to minus and plus
        #if :
        if (ordername != "LO" and
            (plot.usd(dl[ordername][i]) * required_numerical_uncertainty_factor
             > dl[ordername + "_PDF_ERRPLUS"][i] or
             plot.usd(dl[ordername][i]) * required_numerical_uncertainty_factor
             > -dl[ordername + "_PDF_ERRMINUS"][i])):
            rel = plot.unv(dl[ordername][i])
            warnings.warn(
                "too bad numerical precision vs pdf @ " + ordername +
                " num: " + str(plot.usd(dl[ordername][i]) / rel * 100.) +
                "% vs " + str(dl[ordername + "_PDF_ERRPLUS"][i] / rel * 100.) +
                "% to pdf: " +
                str(dl[ordername + "_PDF_ERRMINUS"][i] / rel * 100.) + "%",
                RuntimeWarning)

    mask = dl[ordername + "_PDF_CENTRAL"].notnull()
    dl.loc[mask, ordername + "_PDF"] = unumpy.uarray(
        plot.unv(dl[ordername + "_PDF_CENTRAL"][mask]) +
        dl[ordername + "_PDF_ERRPLUS"][mask] / 2. +
        dl[ordername + "_PDF_ERRMINUS"][mask] / 2.,
        (dl[ordername + "_PDF_ERRPLUS"][mask] -
         dl[ordername + "_PDF_ERRMINUS"][mask]) / 2.)

    return dl


def scale_errors(li, dl, ordernames=None,n_jobs=None):
    """
    Just like `scale_error` but over a list of ordernames.
    """
    if ordernames is None:
        ordernames = ["LO", "NLO", "aNNLO_PLUS_NNLL"]
    r_dl = dl
    for o in ordernames:
        r_dl = scale_error(li, r_dl, o, n_jobs=n_jobs)

    return r_dl

def _scale_error_single(members,i,dl,ordername="LO"):
    if dl["pdfset_nlo"][i] == 0 and dl["mu_f"][i] == 1.0 and dl["mu_r"][
        i] == 1.0:
        scales = []
        ddl = dl[members].drop(columns=["mu_f","mu_r", "precision","max_iters"])
        bol = ddl.eq(ddl.iloc[i]).all(axis='columns')
        for j in range(len(dl["pdfset_nlo"])):
            if bol[j]:
                scales.append(j)
    # index, errplus,errminus
    return (
        i,
        np.max( [plot.unv(dl[ordername][k]) for k in scales]) - plot.unv(dl[ordername][i]),
        np.min( [plot.unv(dl[ordername][k]) for k in scales]) - plot.unv(dl[ordername][i])
    )

def scale_error(li, dl, ordername="LO",n_jobs=None):
    """
    Computes seven-point scale uncertainties from the results where the renormalization and factorization scales are varied by factors of 2 and  relative factors of four are excluded (cf. :meth:`seven_point_scan`).

    Args:
        li (:obj:`list` of :class:`Input`): Input list.
        dl (:obj:`dict`): :class:`Result` dictionary with lists per entry.

    Returns:
        :obj:`dict`: Modified `dl` with new `ordername_{SCALE,SCALE_ERRPLUS,SCALE_ERRMINUS}` entries.
            - `ordername_SCALE` contains a symmetrized :mod:`uncertainties` object.
    """
    global required_numerical_uncertainty_factor
    example = li[0]
    members = [
        attr for attr in dir(example)
        if not callable(getattr(example, attr)) and not attr.startswith("__")
    ]
    dl[ordername + "_SCALE"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl[ordername + "_SCALE_ERRPLUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl[ordername + "_SCALE_ERRMINUS"] = np.array([None] *
                                                 len(dl["pdfset_nlo"]))


    args = [{
        "members": members,
        "i": i,
        "dl": dl,
        "ordername": ordername,
    } for i in range(len(dl["pdfset_nlo"])) if dl["pdfset_nlo"][i] == 0
            and dl["mu_f"][i] == 1.0 and dl["mu_r"][i] == 1.0]
    ret = tpqdm(args,
                _scale_error_single,
                n_jobs=n_jobs if n_jobs is not None else mp.cpu_count(),
                argument_type='kwargs',
                desc="Scale uncertainty @ " + ordername)
    for i,errplus,errminus in ret:
        # lo_unc = pdfset.uncertainty(
        #    [plot.unv(dl["LO"][k]) for k in pdfs], -1)
        dl.loc[i, ordername + "_SCALE_ERRPLUS"] = errplus
        dl.loc[i, ordername + "_SCALE_ERRMINUS"] = errminus
        if (plot.usd(dl[ordername][i]) *
                required_numerical_uncertainty_factor >
                dl[ordername + "_SCALE_ERRPLUS"][i]
                or plot.usd(dl[ordername][i]) *
                required_numerical_uncertainty_factor >
                -dl[ordername + "_SCALE_ERRMINUS"][i]):
            rel = plot.unv(dl[ordername][i])
            warnings.warn(
                "too bad numerical precision vs scale @ num:" + ordername +
                " " + str(plot.usd(dl[ordername][i]) / rel * 100.) +
                "% vs scale:" +
                str(dl[ordername + "_SCALE_ERRPLUS"][i] / rel * 100.) +
                "% to " +
                str(dl[ordername + "_SCALE_ERRMINUS"][i] / rel * 100.) +
                "%", RuntimeWarning)

    mask = dl[ordername + "_SCALE_ERRPLUS"].notnull()
    dl.loc[mask, ordername + "_SCALE"] = unumpy.uarray(
        plot.unv(dl[ordername][mask]) +
        dl[ordername + "_SCALE_ERRPLUS"][mask] / 2. +
        dl[ordername + "_SCALE_ERRMINUS"][mask] / 2.,
        (+dl[ordername + "_SCALE_ERRPLUS"][mask] -
         dl[ordername + "_SCALE_ERRMINUS"][mask]) / 2.)

    return dl


def combine_errors(dl, ordernames=None):
    """
    Just like `combine_error` but over a list of ordernames.
    """
    if ordernames is None:
        ordernames = ["LO", "NLO", "aNNLO_PLUS_NNLL"]
    r_dl = dl
    for o in ordernames:
        r_dl = combine_error(r_dl, o)

    return r_dl


def combine_error(dl: dict, ordername="LO"):
    """
    Combines seven-point scale uncertainties and pdf uncertainties from the results by Pythagorean addition.

    Note:
        Running :func:`scale_errors` and :func:`pdf_errors` before is necessary.

    Args:
        dl (:obj:`dict`): :class:`Result` dictionary with lists per entry.

    Returns:
        :obj:`dict`: Modified `dl` with new `ordername_{COMBINED,ERRPLUS,ERRMINUS}` entries.
            - `ordername_COMBINED` contains a symmetrized :mod:`uncertainties` object.
    """
    dl[ordername + "_NOERR"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl[ordername + "_ERRPLUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl[ordername + "_ERRMINUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl[ordername + "_COMBINED"] = np.array([None] * len(dl["pdfset_nlo"]))

    mask = dl[ordername + "_PDF_CENTRAL"].notnull()
    dl.loc[mask, ordername + "_NOERR"] = plot.unv(dl[ordername +
                                                     ""][mask]).astype(float)
    dl.loc[mask, ordername + "_ERRPLUS"] = np.sqrt(
        dl[ordername + "_PDF_ERRPLUS"][mask].astype(float)**2 +
        dl[ordername + "_SCALE_ERRPLUS"][mask].astype(float)**2)
    dl.loc[mask, ordername + "_ERRMINUS"] = -np.sqrt(
        dl[ordername + "_PDF_ERRMINUS"][mask].astype(float)**2 +
        dl[ordername + "_SCALE_ERRMINUS"][mask].astype(float)**2)
    dl.loc[mask, ordername + "_COMBINED"] = unumpy.uarray(
        plot.unv(dl[ordername + ""][mask]) +
        dl[ordername + "_ERRPLUS"][mask] / 2. +
        dl[ordername + "_ERRMINUS"][mask] / 2.,
        (+dl[ordername + "_ERRPLUS"][mask] - dl[ordername + "_ERRMINUS"][mask])
        / 2.)

    return dl



from typing import List
import warnings
import subprocess
import os
from hepi.input import Input, update_slha
from hepi.run import Runner


class SPhenoRunner(Runner):

    def _check_path(self) -> bool:
        if os.path.exists(os.path.expanduser(self.get_path() + "/bin/SPheno")):
            self.set_path(self.get_path() + "/bin/SPheno")
            return True
        if self.get_path().endswith("SPheno"):
            return True
        return False

    def run(self, slhas: List[Input], **kwargs) -> List[Input]:
        """
        Run the passed list of parameters for SPheno.
    
        Args:
            slhas (:obj:`list` of :class:`Input`): Input parameters with a SLHA file that can be processed by SPheno.
        Returns:
            :obj:`list` of :class:`Input`
        """
        if not self._check_path():
            warnings.warn("The path is not valid for " + self.get_name())
        if os.path.exists("Messages.out"):
            os.remove("Messages.out")
        for s in slhas:
            # Remove Creation time for hash-/caching
            comm = "cp " + self.get_output_dir(
            ) + s.slha + " spheno_tmp.in && " + get_path(
            ) + " spheno_tmp.in && mv " + "SPheno.spc " + self.get_output_dir(
            ) + s.slha + " && sed -i '/Created/d' " + self.get_output_dir(
            ) + s.slha
            proc = subprocess.Popen(comm,
                                    shell=True,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE)
            proc.wait()
            update_slha(s)
        if os.path.exists("Messages.out"):
            with open("Messages.out", 'r') as r:
                t = r.read()
                if t != "":
                    warnings.warn(r.read())
        return slhas


# Backward compatibility
spheno_default_runner = SPhenoRunner("~/git/SPheno-3.3.8/")
"""Default SPheno Runner to provide backward compatibility"""
run = spheno_default_runner.run
set_path = spheno_default_runner.set_path
get_path = spheno_default_runner.get_path



from hepi.results import Result


class NLLFastResult(Result):
    """
    (N)NLL-fast Result Data.
    """

    def __init__(self, lo, nlo, nlo_plus_nll, scale_err_up, scale_err_down,
                 pdf_err_up, pdf_err_down):
        Result.__init__(self, lo, nlo, nlo_plus_nll)
        self.scale_err_up = scale_err_up
        self.scale_err_down = scale_err_down
        self.pdf_err_up = pdf_err_up
        self.pdf_err_down = pdf_err_down



import warnings
from hepi.input import Order
from .. import Input, Result
import re
from uncertainties import ufloat_fromstr


class MadgraphResult(Result):
    """ MadGraph Result Data."""

    def __init__(self, lo, nlo):
        """Sets LO and NLO result. NLO+NLL is set to None."""
        Result.__init__(self, lo, nlo, None)


def is_valid(file: str, p: Input, d) -> bool:
    """
    Verifies that an file is a complete output.

    Args:
        file (str): File path to be parsed.
        p (:class:`hepi.Input`): Input parameters.
        d (:obj:`dict`): Param dictionary.

    Returns:
        bool : True if `file` could be parsed.
    """
    order = p.order
    res = parse_single(file)
    if order is not Order.NLO and order is not Order.LO:
        warnings.warn("MadGraph has only NLO and LO computation.")
    if res.LO is not None and order is Order.LO:
        return True
    if res.LO is not None and res.NLO is not None and order is Order.NLO:
        return True
    print("RESTART", res.LO, res.NLO, res.NLO_PLUS_NLL, file)
    return False


def parse_single(file) -> MadgraphResult:
    """
    Extracts Result from MadGraph output file.

    Note:
        This is only the result of one order. Therefore LO and NLO result in the return value are the same.


    Args:
        file (str): File path to be parsed.

    Returns:
        :class:`MadGraphResult` : If a value is not found in the file None is used.

    """
    # TODO generalize units like RS
    lo_pattern = re.compile(r'^\s*Total cross section:\s(\S+.*) pb')

    lo_result = None
    nlo_result = None
    with open(file) as output:
        for line in output:
            tmp = lo_pattern.search(line)
            if tmp is not None:
                if lo_result is None:
                    lo_result = ufloat_fromstr(
                        tmp.group(1).replace("+-", "+/-"))
                elif nlo_result is None:
                    nlo_result = ufloat_fromstr(
                        tmp.group(1).replace("+-", "+/-"))

    return MadgraphResult(lo_result, nlo_result)



from concurrent.futures import ThreadPoolExecutor
from logging import warning
import os
import subprocess
from subprocess import Popen, PIPE
from typing import List
import warnings

import numpy as np
from hepi.input import Input, Order, get_input_dir, get_output_dir, get_pre
from hepi.results import Result
from hepi.util import DL2DF, LD2DL, DictData, namehash
from smpl.parallel import par
import time
import tqdm
from pqdm.threads import pqdm as tpqdm
from pqdm.processes import pqdm as ppqdm
import multiprocessing as mp
#from pqdm.processes import pqdm

class RunParam(DictData):
    """Abstract class that is similar to a dictionary but with fixed keys."""

    def __init__(self,
                 skip: bool = False,
                 in_file: str = None,
                 out_file: str = None,
                 execute: str = None,
                 name: str = None):
        self.name = name
        self.skip = skip
        self.in_file = in_file
        self.out_file = out_file
        self.execute = execute


class Runner:

    def __init__(self,
                 path: str,
                 in_dir: str = None,
                 out_dir: str = None,
                 pre=None):
        self.path = path
        if in_dir is None:
            self.in_dir = get_input_dir()
        else:
            self.in_dir = in_dir
        if out_dir is None:
            self.out_dir = get_output_dir()
        else:
            self.out_dir = out_dir
        if pre is None:
            self.pre = get_pre()
        else:
            self.pre = pre

    def orders(self) -> List[Order]:
        """List of supported Orders in this runner."""
        return [e.value for e in Order]

    def get_name(self) -> str:
        """Returns the name of the runner."""
        return type(self).__name__

    def get_version(self) -> str:
        return "?"

    def _sub_run(self, coms: List[str]) -> str:
        process = Popen(coms, stdout=PIPE)
        (output, err) = process.communicate()
        exit_code = process.wait()
        if exit_code != 0:
            return err.decode()
        else:
            return output.decode()

    def _check_path(self) -> bool:
        """Checks if the passed path is valid."""
        return True

    def _prepare(self, p: Input, skip=False, assume_valid = False, **kwargs) -> RunParam:
        skip_ = skip
        d = p.__dict__
        d["runner"] = str(type(self).__name__) + "-" + self.get_version(
        )  # TODO re add version, but removed for reusable hashing!
        name = namehash("_".join("".join(str(_[0]) + "_" + str(_[1]))
                                 for _ in d.items()).replace("/", "-"))
        #print(name)
        skip = False
        if skip_ and os.path.isfile(self.get_output_dir() + name +
                                    ".out") and (assume_valid or self._is_valid(
                                        self.get_output_dir() + name + ".out",
                                        p,
                                        d,
                                        skip=skip_,
                                        **kwargs)):
            #print(".", end='')
            skip = True
        else:
            #print('|', end='')
            pass
        return RunParam(execute=self.get_output_dir() + name + ".sh",
                        in_file=self.get_output_dir() + name + ".in",
                        out_file=self.get_output_dir() + name + ".out",
                        skip=skip,
                        name=name)

    def _check_input(self, param: Input, **kwargs) -> bool:
        if param.order not in self.orders():
            warnings.warn("Order " + str(param.order) + " not supported in " +
                          type(self).__name__)
            return False
        return True

    def _prepare_all(self,
                     params: List[Input],
                     skip=True,
                     n_jobs = None,
                     **kwargs) -> List[RunParam]:
        """
        Prepares all parameters for execution.

        Args:
            params (List[:class:`hepi.Input`]): List of input parameters.
            skip (bool, optional): If True, the runner will check if the output file already exists and skip the execution if it does. Defaults to True.
            n_jobs (int): Number of parallel jobs. If None, use all available cores.
        """
        ret = []

        #ret = my_parallel(self._check_input,params,desc="Checking input")
        ret = tpqdm(params,
                    self._check_input,
                    n_jobs=n_jobs if n_jobs is not None else mp.cpu_count(),
                    desc="Checking input")
        if not np.alltrue(ret):
            warnings.warn("Check input failed.")
            return []
        #ret = my_parallel(
        #    lambda p: self._prepare(p, skip=skip, **kwargs),
        #    params,
        #    #n_jobs=mp.cpu_count(),
        #    desc="Preparing")
        args = [{'p': p, 'skip': skip, **kwargs} for p in params]
        ret = ppqdm(args,
                    self._prepare,
                    n_jobs=n_jobs if n_jobs is not None else mp.cpu_count(),
                    argument_type='kwargs',
                    desc="Preparing")
        skipped = 0
        not_skipped = 0
        for r in ret:
            if r.skip:
                skipped += 1
            else:
                not_skipped += 1
        print("Skipped: " + str(skipped) + " Not skipped: " + str(not_skipped))
        #for p in params:
        #    if not self._check_input(p):
        #        warnings.warn("Check input failed.")
        #        return []
        #    ret.append(self._prepare(p, skip=skip, **kwargs))
        return ret

    def run(self,
            params: List[Input],
            skip=True,
            parse=True,
            parallel=True,
            sleep=0,
            run=True,
            ignore_error=False,
            n_jobs = None,
            **kwargs):
        """
		Run the passed list of parameters.

		Args:
		    params (:obj:`list` of :class:`hepi.Input`): All parameters that should be executed/queued.
		    skip (bool): True means stored runs will be skipped. Else the are overwritten.
		    parse (bool): Parse the results. 
		        This is not the prefered cluster/parallel mode, as there the function only queues the job.
		    parallel (bool): Run jobs in parallel.
		    sleep (int): Sleep seconds after starting job.
            run (bool): Actually start/queue runner.
            ignore_error (bool): Continue instead of raising Exceptions. Also ignores hash collisions.
            n_jobs (int): Number of parallel jobs. If None, use all available cores.

		Returns:
		    :obj:`pd.DataFrame` : combined dataframe of results and parameters. The dataframe is empty if `parse` is set to False.

		"""
        if not self._check_path():
            warnings.warn("The path is not valid for " + self.get_name())
            if not ignore_error:
                raise RuntimeError("The path is not valid for " +
                                   self.get_name())
        rps = self._prepare_all(params,
                                parse=parse,
                                skip=skip,
                                ignore_error=ignore_error,
                                n_jobs=n_jobs,
                                **kwargs)
        #print("= " + str(len(params)) + " jobs")
        if sleep is None:
            sleep = 0 if parse else 5
        if run:
            self._run(rps,
                      wait=parse,
                      parallel=parallel,
                      sleep=sleep,
                      n_jobs = n_jobs,
                      **kwargs)
        if parse:
            outs = LD2DL(rps)["out_file"]
            results = self.parse(outs,n_jobs=n_jobs)
            rdl = LD2DL(results)
            pdl = LD2DL(params)
            return DL2DF({**rdl, **pdl})
        return DL2DF({})

    def _run(self,
             rps: List[RunParam],
             wait=True,
             parallel=True,
             sleep=0,
             n_jobs = None,
             **kwargs):
        """
		Runs Runner per :class:`RunParams`.
	
		Args:
		    rps (:obj:`list` of :class:`RunParams`): Extended run parameters.
		    bar (bool): Enable info bar.
		    wait (bool): Wait for parallel runs to finish.
		    sleep (int): Sleep seconds after starting subprocess.
		    parallel (bool): Run jobs in parallel.
            n_jobs (int): Number of parallel jobs. If None, use all available cores.
	
		Returns:
		    :obj:`list` of int: return codes from jobs if `no_parse` is False.
		"""
        # get cluster or niceness prefix
        template = self.get_pre() + " " + "{}"

        # Run commands in parallel
        processes = []
        cmds = []

        for rp in rps:
            if not rp.skip:
                command = template.format(rp.execute)
                cmds.append(command)

        if wait and parallel:
            return tpqdm(cmds,
                  lambda c: subprocess.call(c,shell=True),
                  n_jobs=n_jobs if n_jobs is not None else mp.cpu_count(),
                  desc="Running")
            #with ThreadPoolExecutor(max_workers=n_jobs if n_jobs is not None else mp.cpu_count()) as executor:
            #    for cmd in cmds:
            #        processes.append(executor.submit(subprocess.call, cmd, shell=True))
            #        time.sleep(sleep)
            #    return [p.result() for p in processes]

        for command in cmds:
            command = template.format(rp.execute)
            process = subprocess.Popen(command, shell=True)
            processes.append(process)
            if not parallel:
                process.wait()
            # Forced delay to prevent overloading clusters when registering jobs
            time.sleep(sleep)

        if wait:
            # Collect statuses
            output = [p.wait() for p in processes]
            return output
        return []

    def _is_valid(self, file: str, p: Input, d, **kwargs) -> bool:
        """
		Verifies that a file is a complete output.
	
		Args:
		    file (str): File path to be parsed.
		    p (:class:`hepi.Input`): Onput parameters.
		    d (:obj:`dict`): Param dictionary.
	
		Returns:
		    bool : True if `file` could be parsed.
		"""
        res = self._parse_file(file)
        if res.LO is None and p.order is Order.LO:
            return False
        if res.NLO is None and p.order is Order.NLO:
            return False
        if res.NLO_PLUS_NLL is None and p.order is Order.NLO_PLUS_NLL:
            return False
        if res.aNNLO_PLUS_NNLL is None and p.order is Order.aNNLO_PLUS_NNLL:
            return False
        return True

    def parse(self, outputs: List[str],n_jobs=None) -> List[Result]:
        """
		Parses Resummino output files and returns List of Results.
	
		Args:
		    outputs (:obj:`list` of `str`): List of the filenames to be parsed.
            n_jobs (int): Number of parallel jobs. If None, use all available cores.
	
		Returns:
		    :obj:`list` of :class:`hepi.resummino.result.ResumminoResult`
	
		"""
        rsl = []
        #for r in parallel(self._parse_file, outputs):
        #    rsl.append(r)

        # parallelized opens to many files at times
        #rsl = my_parallel(self._parse_file, outputs, desc="Parsing")
        rsl = tpqdm(outputs,
                   self._parse_file,
                   n_jobs=n_jobs if n_jobs is not None else mp.cpu_count(),
                   desc="Parsing")
        return rsl
        #for o in tqdm.tqdm(outputs):
        #    rsl.append(self._parse_file(o))
        #return rsl

    def _parse_file(self, file: str) -> Result:
        """
		Extracts results from an output file.

		Args:
		    file (str): File path to be parsed.

		Returns:
		    :class:`Result` : If a value is not found in the file None is used.

		"""
        return None

    def get_path(self) -> str:
        """
		Get the Runner path.

		Returns:
		    str: current Runner path.
		"""
        return self.path

    def get_input_dir(self) -> str:
        """
		Get the input directory.

		Returns:
		    str: :attr:`in_dir`
		"""
        return self.in_dir

    def get_output_dir(self) -> str:
        """
		Get the input directory.

		Returns:
		    str: :attr:`out_dir`
		"""
        return self.out_dir

    def get_pre(self) -> str:
        """
		Gets the command prefix.

		Returns:
		    str: :attr:`pre`
		"""
        return self.pre

    def set_path(self, p: str):
        """
		Set the path to the Runner folder containing the binary in './bin' or './build/bin'.

		Args:
		    p (str): new path.
		"""
        if os.path.isdir(p):
            self.path = p + ("/" if p[-1] != "/" else "")
        self.path = p

    def set_input_dir(self, indir: str):
        """
		Sets the input directory.

		Args:
		    indir (str): new input directory.
		"""
        self.in_dir = indir

    def set_output_dir(self, outdir: str, create: bool = True):
        """
		Sets the output directory.

		Args:
		    outdir (str): new output directory.
			create (bool): create directory if not existing.
		"""
        if create:
            os.makedirs(outdir, exist_ok=True)
        self.out_dir = outdir

    def set_pre(self, ppre: str):
        """
		Sets the command prefix.

		Args:
		    ppre (str): new command prefix.
		"""
        self.pre = ppre



from matplotlib.image import NonUniformImage
from collections.abc import Iterable
import matplotlib as mpl
from smpl import plot as splot
import numpy as np
from scipy.interpolate import make_interp_spline, BSpline

import matplotlib.pyplot as plt

import pyslha

import matplotlib.cm as cm
from matplotlib import colors

from .input import Input, get_output_dir, replace_macros
from .util import get_name
from matplotlib.ticker import NullFormatter
from scipy import integrate as auc


def title(i: Input,
          axe=None,
          scenario=None,
          diff_L_R=None,
          extra="",
          cms_energy=True,
          pdf_info=True,
          id=False,
          **kwargs):
    """Sets the title on axis `axe`."""
    if axe is None:
        axe = plt.gca()
    axe.set_title("$pp\\to" + get_name(i.particle1) + get_name(i.particle2) +
                  "$" + (" at $\\sqrt{S} = " + str(i.energy / 1000) +
                         "$ TeV" if cms_energy else "") +
                  ((" for " +
                    i.slha.split(".")[0]) if scenario is None else scenario) +
                  (" with " + i.pdf_nlo if pdf_info else "") + " " + extra +
                  ((" [" + i.id + "]") if id else ""))


def energy_plot(dict_list,
                y,
                yscale=1.,
                xaxis="E [GeV]",
                yaxis="$\\sigma$ [pb]",
                label=None,
                **kwargs):
    """Plot energy on the x-axis."""
    plot(dict_list,
         "energy",
         y,
         label=label,
         xaxis=xaxis,
         yaxis=yaxis,
         logy=True,
         yscale=yscale,
         **kwargs)


def mass_plot(dict_list,
              y,
              part,
              logy=True,
              yaxis="$\\sigma$ [pb]",
              yscale=1.,
              label=None,
              xaxis = None,
              **kwargs):
    dict_list["mass_" + str(part)] = get_mass(dict_list, abs(part))
    if xaxis is None:
        xaxis = "$m_{" + get_name(part) + "}$ [GeV]"
    plot(dict_list,
         "mass_" + str(part),
         y,
         label=label,
         yaxis=yaxis,
         xaxis=xaxis,
         logy=logy,
         yscale=yscale,
         **kwargs)


def mass_vplot(dict_list,
               y,
               part,
               logy=True,
               yaxis="$\\sigma$ [pb]",
               yscale=1.,
               label=None,
               mask=None,
               **kwargs):
    vplot(get_mass(dict_list, part)[mask],
          y[mask],
          label=label,
          xaxis="$m_{" + get_name(part) + "}$ [GeV]",
          yaxis=yaxis,
          logy=logy,
          yscale=yscale,
          mask=mask,
          **kwargs)


def get_mass(l: dict, iid: int):
    """
    Get the mass of particle with id `iid` out of the list in the "slha" element in the dict.

    Returns
        :obj:`list` of float : masses of particles in each element of the dict list.

    """
    ret = []
    for s in l["slha"]:
        d = pyslha.read(get_output_dir() + s)
        ret.append(d.blocks["MASS"][abs(iid)])
    return np.array(ret)


def plot(dict_list,
         x,
         y,
         label=None,
         xaxis="E [GeV]",
         yaxis="$\\sigma$ [pb]",
         ratio=False,
         K=False,
         K_plus_1=False,
         logy=True,
         yscale=1.,
         mask=None,
         **kwargs) -> None:
    """
    Creates a plot based on the entries `x`and `y` in `dict_list`.
    
    Examples

    .. plot::
        :include-source:

        >>> import urllib.request
        >>> import hepi 
        >>> dl = hepi.load(urllib.request.urlopen(
        ... "https://raw.githubusercontent.com/fuenfundachtzig/xsec/master/json/pp13_hino_NLO%2BNLL.json"
        ... ))
        >>> hepi.plot(dl,"N1","NLO_PLUS_NLL",xaxis="$m_{\\\\tilde{\\\\chi}_1^0}$ [GeV]")
    """
    if isinstance(y, Iterable) and not isinstance(y, str):
        for yi in y:
            plot(dict_list, x, yi, label, xaxis, yaxis, ratio, K, K_plus_1,
                 logy, yscale, mask, **kwargs)
        return
    # TODO use kwargs
    if label is None:
        label = y.replace("_PLUS_", "+").replace("_OVER_", "/")
    if label == "":
        label = None

    vx = dict_list[x].to_numpy()[mask]
    vy = dict_list[y].to_numpy()[mask]

    if K:
        yaxis = "$K$"
        yscale = 1.0
        vy = vy / splot.unv(dict_list["LO"].to_numpy()[mask])
        if K_plus_1:
            vy = vy + vy / vy
    if ratio:
        yaxis = "Ratio"
        yscale = 1.0
        vy = vy / splot.unv(vy)

    vplot(vx, vy, label, xaxis, yaxis, logy, yscale, mask=mask, **kwargs)


def index_open(var, idx):
    if len(idx) == 1:
        return var[idx]
    return index_open(var[idx[0]], idx[1:])


def slha_data(li, index_list):
    vx = []
    for l in li:
        b = pyslha.read(get_output_dir() + l.slha)
        vx.append(index_open(b.blocks, index_list))
    return np.array(vx)


def slha_plot(li, x, y, **kwargs):
    vx = slha_data(li, x)
    vy = slha_data(li, y)

    vplot(np.array(vx), np.array(vy), **kwargs)


def vplot(x,
          y,
          label=None,
          xaxis="E [GeV]",
          yaxis="$\\sigma$ [pb]",
          logy=True,
          yscale=1.,
          interpolate=True,
          plot_data=True,
          data_color=None,
          mask=-1,
          fill=False,
          data_fmt=".",
          fmt="-",
          print_area=False,
          sort=True,
          **kwargs):
    """
    Creates a plot based on the values in `x`and `y`.
    
    """
    color = data_color
    if label is None:
        #label = "??"
        pass
    if mask is None:
        x = x[0]
        y = y[0]
    if sort:
        permute = x.argsort(kind='stable')
        vx = x[permute]
        vy = y[permute]
    else:
        vx = x
        vy = y

    if data_color is None and 'next_color' not in kwargs:
        if 'axes' in kwargs and kwargs['axes'] is not None:
            bl, = kwargs['axes'].plot([], [])
        else:
            bl, = plt.gca().plot([], [])
        color = bl.get_color()
    iii = splot.data(vx,
                     vy * yscale,
                     label=label,
                     xaxis=xaxis,
                     yaxis=yaxis,
                     logy=logy,
                     data_color=color,
                     fit_color=color,
                     also_data=plot_data,
                     interpolate=interpolate,
                     fmt=data_fmt,
                     sigmas=0 if not fill else 1,
                     **kwargs)
    if iii is not None:
        #ii = iii[0]
        ix = iii[1]
        iy = iii[2]
    if print_area:
        print('computed AUC using scipy.integrate.simpson: {}'.format(auc.simpson(iy, ix)))
    if ((np.any(np.less(vy, 0)) or (interpolate and np.any(np.less(iy, 0))))
            and logy):
        splot.data(vx,
                   -vy * yscale,
                   label="-" + label,
                   xaxis=xaxis,
                   yaxis=yaxis,
                   logy=logy,
                   data_color=color,
                   fit_color=color,
                   also_data=plot_data,
                   interpolate=interpolate,
                   interpolate_fmt=(0, (3, 1, 3, 1, 1, 1)),
                   fmt=data_fmt,
                   sigmas=0 if not fill else 1,
                   **kwargs)


def mass_mapplot(dict_list,
                 part1,
                 part2,
                 z,
                 logz=True,
                 zaxis="$\\sigma$ [pb]",
                 zscale=1.,
                 label=None):
    mapplot(dict_list,
            "mass_" + str(part1),
            "mass_" + str(part2),
            z,
            xaxis="$M_{" + get_name(part1) + "}$ [GeV]",
            yaxis="$M_{" + get_name(part2) + "}$ [GeV]",
            zaxis=zaxis,
            logz=logz,
            zscale=zscale)


def mapplot(dict_list, x, y, z, xaxis=None, yaxis=None, zaxis=None, **kwargs):
    """

    Examples

    .. plot::
        :include-source:

        >>> import urllib.request
        >>> import hepi 

        >>> dl = hepi.load(urllib.request.urlopen(
        ... "https://raw.githubusercontent.com/APN-Pucky/xsec/master/json/pp13_SGmodel_GGxsec_NLO%2BNLL.json"
        ... ),dimensions=2)
        >>> hepi.mapplot(dl,"gl","sq","NLO_PLUS_NLL",xaxis="$m_{\\\\tilde{g}}$ [GeV]",yaxis="$m_{\\\\tilde{q}}$ [GeV]" , zaxis="$\\\\sigma_{\\\\mathrm{NLO+NLL}}$ [pb]")
    """
    if xaxis is None:
        xaxis = x
    if yaxis is None:
        yaxis = y
    if zaxis is None:
        zaxis = replace_macros(z)
    vx = dict_list[x]
    vy = dict_list[y]
    vz = dict_list[z]
    splot.plot2d(vx,
                 vy,
                 vz,
                 style="pcolormesh",
                 xaxis=xaxis,
                 yaxis=yaxis,
                 zaxis=zaxis,
                 **kwargs)


map_vplot = lambda *a, **da: splot.plot2d(*a, style="pcolormesh", **da)
scatter_vplot = lambda *a, **da: splot.plot2d(*a, style="scatter", **da)


def scatterplot(dict_list,
                x,
                y,
                z,
                xaxis=None,
                yaxis=None,
                zaxis=None,
                **kwargs):
    """
    Scatter map 2d.
    Central color is the central value, while the inner and outer ring are lower and upper bounds of the uncertainty interval.

    Examples

    .. plot::
        :include-source:

        >>> import urllib.request
        >>> import hepi 
        >>> dl = hepi.load(urllib.request.urlopen(
        ... "https://raw.githubusercontent.com/APN-Pucky/xsec/master/json/pp13_hinosplit_N2N1_NLO%2BNLL.json"
        ... ),dimensions=2)
        >>> hepi.scatterplot(dl,"N1","N2","NLO_PLUS_NLL",xaxis="$m_{\\\\tilde{\\\\chi}_1^0}$ [GeV]",yaxis="$m_{\\\\tilde{\\\\chi}_2^0}$ [GeV]" , zaxis="$\\\\sigma_{\\\\mathrm{NLO+NLL}}$ [pb]")

    """
    if xaxis is None:
        xaxis = x
    if yaxis is None:
        yaxis = y
    if zaxis is None:
        zaxis = replace_macros(z)
    vx = dict_list[x]
    vy = dict_list[y]
    vz = dict_list[z]
    splot.plot2d(vx,
                 vy,
                 vz,
                 style="scatter",
                 xaxis=xaxis,
                 yaxis=yaxis,
                 zaxis=zaxis,
                 **kwargs)


fig = None
axs = None
lines = []
labels = []


def err_plt(axes, x, y, label=None, error=False):
    v = label
    ind = np.argsort(splot.unv(x), kind='stable')
    if error:
        l, _, _ = axes.errorbar(x.to_numpy()[ind],
                                splot.unv(y)[ind],
                                yerr=splot.usd(y),
                                capsize=5,
                                label=v)
        return l
    else:
        l = axes.plot(x.to_numpy()[ind], splot.unv(y)[ind], label=v)
        return l[0]


def scale_plot(dict_list,
               vl,
               seven_point_band=False,
               cont=False,
               error=True,
               li=None,
               plehn_color=False,
               yscale=1.,
               unit="pb",
               yaxis=None,
               **kwargs):
    """Creates a scale variance plot with 5 panels (xline)."""
    global fig, axs, lines, labels
    cycle_safe = mpl.rcParams['axes.prop_cycle']
    if plehn_color:
        mpl.rcParams['axes.prop_cycle'] = mpl.cycler(color=["b", "r", "g"])
    if not cont:
        fig, axs = plt.subplots(1, 5, figsize=(12, 3), sharey=True)
        # Remove horizontal space between axes
        fig.subplots_adjust(wspace=0)
        if li is not None:
            title(axs[2], li[0], **kwargs)

    mr = dict_list["mu_r"]
    mf = dict_list["mu_f"]

    if not cont:
        lines = []
        labels = []
        axs[0].plot([], [], ' ', color='k', label="$\\mu_R=" + "\\mu_F$")
        axs[1].plot([], [],
                    ' ',
                    color='k',
                    label="$\\mu_R=" + str(np.max(mr)) + "\\mu_0$")
        axs[2].plot([], [],
                    ' ',
                    color='k',
                    label="$\\mu_F=" + str(np.min(mf)) + "\\mu_0$")
        axs[3].plot([], [],
                    ' ',
                    color='k',
                    label="$\\mu_R=" + str(np.min(mr)) + "\\mu_0$")
        axs[4].plot([], [],
                    ' ',
                    color='k',
                    label="$\\mu_F=" + str(np.max(mf)) + "\\mu_0$")

    for v in vl:
        mv = dict_list[v] * yscale
        if seven_point_band:
            #mask = (mf/mr < 4.) & (mf/mr > 1./4.) & (mf <= 2) & (mf >= 1./2.) & (mr <= 2) & (mr >= 1./2.)
            mask = (((mf == 2.) & (mr == 2.)) | ((mf == 2.) & (mr == 1.)) |
                    ((mf == 1.) & (mr == 1.)) | ((mf == 1.) & (mr == 2.)) |
                    ((mf == 1 / 2.) & (mr == 1 / 2.)) |
                    ((mf == 1 / 2.) & (mr == 1.)) | ((mf == 1.) &
                                                     (mr == 1 / 2.)))
            mvmax = np.max(splot.unv(mv[mask]))
            mvmin = np.min(splot.unv(mv[mask]))

        mask = mf == mr
        l = err_plt(axs[0], mf[mask], mv[mask], error=error)
        #l, _, _ = axs[0].errorbar(mf[mask], splot.unv(mv[mask]),
        #                          yerr=splot.usd(mv[mask]), capsize=5, label=v)
        if seven_point_band:
            axs[0].fill_between(mf[mask],
                                mvmax,
                                mvmin,
                                facecolor=l.get_color(),
                                alpha=0.3)

        mask = mr == np.max(mr)
        l = err_plt(axs[1], mf[mask], mv[mask], error=error)
        lines.append(l)
        labels.append("$\\sigma_{\\mathrm{" +
                      v.replace("_PLUS_", "+").replace(" ", "\\ ") + "} }$")
        #l, _, _ = axs[1].errorbar(mf[mask], splot.unv(mv[mask]),
        #                          yerr=splot.usd(mv[mask]), capsize=5)
        if seven_point_band:
            axs[1].fill_between(mf[mask],
                                mvmax,
                                mvmin,
                                facecolor=l.get_color(),
                                alpha=0.3)

        mask = mf == np.min(mf)
        l = err_plt(axs[2], mr[mask], mv[mask], error=error)
        #l, _, _ = axs[2].errorbar(mr[mask], splot.unv(mv[mask]),
        #                          yerr=splot.usd(mv[mask]), capsize=5)
        if seven_point_band:
            axs[2].fill_between(mr[mask],
                                mvmax,
                                mvmin,
                                facecolor=l.get_color(),
                                alpha=0.3)

        mask = mr == np.min(mr)
        l = err_plt(axs[3], mf[mask], mv[mask], error=error)
        #l, _, _ = axs[3].errorbar(mf[mask], splot.unv(mv[mask]),
        #                          yerr=splot.usd(mv[mask]), capsize=5)
        if seven_point_band:
            axs[3].fill_between(mf[mask],
                                mvmax,
                                mvmin,
                                facecolor=l.get_color(),
                                alpha=0.3)

        mask = mf == np.max(mf)
        l = err_plt(axs[4], mr[mask], mv[mask], error=error)
        #l, _, _ = axs[4].errorbar(mr[mask], splot.unv(mv[mask]),
        #                          yerr=splot.usd(mv[mask]), capsize=5)
        if seven_point_band:
            f = axs[4].fill_between(mr[mask],
                                    mvmax,
                                    mvmin,
                                    facecolor=l.get_color(),
                                    alpha=0.3)
            lines.append(f)
            labels.append("$\\Delta \\sigma_{\\mathrm{" + v.replace(
                "NLO_PLUS_NLL", "NLO+NLL").replace(" ", "\\<space>") + "} }$")

    axs[0].set_ylabel("$\\sigma$ [" + unit + "]" if yaxis is None else yaxis)

    axs[0].set_xscale("log")
    axs[0].set_xlim(np.min(mf), np.max(mf))
    axs[0].set_xlabel("$\\mu_{R,F}/\\mu_0$")

    # axs[1].plot(t, s2)
    axs[1].set_xscale("log")
    axs[1].set_xlim(np.max(mf), np.min(mf))
    axs[1].set_xticks([1.])
    axs[1].xaxis.set_minor_formatter(NullFormatter())
    axs[1].set_xlabel("$\\mu_{F}/\\mu_0$")

    # axs[2].plot(t, s3)
    axs[2].set_xscale("log")
    axs[2].set_xlim(np.max(mf), np.min(mf))
    axs[2].set_xticks([1.])
    axs[2].xaxis.set_minor_formatter(NullFormatter())
    axs[2].set_xlabel("$\\mu_{R}/\\mu_0$")

    # axs[3].plot(t, s3)
    axs[3].set_xscale("log")
    axs[3].set_xlim(np.min(mf), np.max(mf))
    axs[3].set_xticks([1.])
    axs[3].xaxis.set_minor_formatter(NullFormatter())
    axs[3].set_xlabel("$\\mu_{F}/\\mu_0$")

    # axs[4].plot(t, s3)
    axs[4].set_xscale("log")
    axs[4].set_xlim(np.min(mf), np.max(mf))
    axs[4].set_xticks([1.])
    axs[4].xaxis.set_minor_formatter(NullFormatter())
    axs[4].set_xlabel("$\\mu_{R}/\\mu_0$")

    axs[0].legend(handletextpad=-0.0, handlelength=0, fancybox=False)
    axs[1].legend(handletextpad=-0.0, handlelength=0, fancybox=False)
    axs[2].legend(handletextpad=-0.0, handlelength=0, fancybox=False)
    axs[3].legend(handletextpad=-0.0, handlelength=0, fancybox=False)
    axs[4].legend(handletextpad=-0.0, handlelength=0, fancybox=False)
    fig.legend()
    fig.legends = []
    fig.legend(handles=lines, labels=labels, loc="center right")
    plt.subplots_adjust(right=0.84)
    # plt.show()
    mpl.rcParams['axes.prop_cycle'] = cycle_safe


def central_scale_plot(dict_list,
                       vl,
                       cont=False,
                       error=True,
                       yscale=1.,
                       unit="pb",
                       yaxis=None):
    """Creates a scale variance plot with 3 panels (ystacked)."""
    global fig, axs
    if not cont:
        fig, axs = plt.subplots(3, 1, figsize=(12, 8), sharex=True)
        # Remove horizontal space between axes
        fig.subplots_adjust(hspace=0)

    mr = dict_list["mu_r"]
    mf = dict_list["mu_f"]

    for v in vl:
        mv = dict_list[v] * yscale

        mask = mf == mr
        l = err_plt(axs[0], mf[mask], mv[mask], label=v, error=error)

        mask = mf == 1.0
        l = err_plt(axs[1], mr[mask], mv[mask], error=error)

        mask = mr == 1.0
        l = err_plt(axs[2], mf[mask], mv[mask], error=error)

    if not cont:
        axs[0].plot([], [], ' ', label="$\\mu_R=\\mu_F=\\mu$")
        axs[1].plot([], [], ' ', label="$\\mu_F=\\mu_0$, $\\mu_R=\\mu$")
        axs[2].plot([], [], ' ', label="$\\mu_R=\\mu_0$, $\\mu_F=\\mu$")

    axs[1].set_ylabel("$\\sigma$ [" + unit + "]" if yaxis is None else yaxis)

    axs[0].set_xscale("log")
    #axs[0].set_xlim(np.min(mf), np.max(mf))
    axs[2].set_xlabel("$\\mu/\\mu_0$")

    axs[0].legend()
    axs[1].legend()
    axs[2].legend()
    # plt.show()


def init_double_plot(figsize=(6, 8),
                     sharex=True,
                     sharey=False,
                     gridspec_kw={'height_ratios': [3, 1]}):
    """Initialze subplot for Ratio/K plots with another figure below."""
    fig, axs = plt.subplots(2,
                            1,
                            figsize=figsize,
                            sharex=sharex,
                            sharey=sharey,
                            gridspec_kw=gridspec_kw)
    # Remove horizontal space between axes
    fig.subplots_adjust(hspace=0)
    return fig, axs
